// cBackground.cpp

#include "cBackground.h"

#include <GL/gl.h>

#include <cstdio>
#include <memory.h>
#include <stdlib.h>

using namespace std;

#include "psi3d/macros.h"

cDomeBackground::cDomeBackground() {
    unsigned int texname;
    const int WIDTH = 64;
    const int HEIGHT = 64;
    const int DEPTH = 1;
    unsigned char texels[WIDTH * HEIGHT * DEPTH * 3];
    float s = 3;
    generateTexture3D(WIDTH, HEIGHT, DEPTH, texels, deltaFunc_NoisyPlasma, ColorFunc_Delta, ColorFunc_Delta, ColorFunc_Delta, 1 * s, 1 * s, 1 * s);
    glBindTexture2D(texname, true, false, true, true, WIDTH, HEIGHT, texels);
    mTextures[0] = texname;
}

void cDomeBackground::drawBackground(float hour) {
    // Push random seed because the current seed is going to be
    // replaced by a (deterministic) constant value for clouds.
    unsigned int seed = rand();
    
    // hour = [0,24]
    // => [-90,   0, 90, 180, 270]
    // => [ -1,   0,  1,   0,  -1] = sin
    // => [  0, 0.5,  1, 0.5,   0] = s brightness sine-wave.
    float t = ((hour * 360.0 / 24.0) - 90.0); // = [-90,+270]
    float s = sin(t * 0.017454)*0.5 + 0.5; // = [0,1]

    // Set gradient colors according to brightness sine-wave.
    vectorset(mUpColor, 0.0, 0.0, s * 0.8 + 0.2);
    vectorset(mLoColor, s * 0.95 + 0.05, s * 0.95 + 0.05, s * 0.95 + 0.05);
    vectorset(mBtColor, 0.4 * s + 0.1, 0.2 * s + 0.1, 0.1 * s + 0.1);

    if (1) {
        float p[] = {700, 100, 0, 1};
        float a[] = {0.3, 0.3, 0.3, 1};
        glLightfv(GL_LIGHT0, GL_POSITION, p);
        glLightfv(GL_LIGHT0, GL_AMBIENT, a);
        glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
        glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0001);
        glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.000001);
        glEnable(GL_NORMALIZE);
    }
    
    GLint mode;
    glGetIntegerv(GL_POLYGON_MODE, &mode);
    if (mode == GL_FILL) {
        drawUpperDome();
        drawLowerDome();
        drawMountains();
        drawClouds();
        drawGround();
    } else {
        drawMountains();
    }

    // Pop a seed.
    srand(seed);
}

void cDomeBackground::drawGround() {
    //return;
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glDisable(GL_FOG);
        glBindTexture(GL_TEXTURE_2D, mTextures[0]);

        // Get current Camera-Matrix.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        // Position from 4th Col.
        float p[3];
        loop3i(p[i] = m[12 + i]);
        // Reset Position on 4th Col.
        loop3i(m[12 + i] = 0);
        // Apply transpose-inverse to position.
        float n[16];
        memcpy(n, m, sizeof(float)*16);
        transposeMatrix4(m);
        applyMatrix2(m, p);
        //printf("height over ground: %.2f\n", p[1]);
        //printf("Position: %.2f %.2f\n", p[0], p[2]);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(n); // Transpose to undo previous transpose.
            glTranslatef(0, p[Y], 0);

            float color2[3] = {0.4, 0.2, 0.1};
            memcpy(color2, mBtColor, sizeof (float) *3);

            const float tdiv = 5;
            const float radius = 200;

            float dx = -p[0] / tdiv;
            float dz = -p[2] / tdiv;

            glBegin(GL_TRIANGLE_FAN);
            glColor4f(mBtColor[0], mBtColor[1], mBtColor[2], 1.0f);
            glNormal3f(0, 1, 0);
            glTexCoord2f(dx, dz);
            glVertex3f(0.0, 0.0, 0.0);
            glColor4f(mBtColor[0], mBtColor[1], mBtColor[2], 0.0f);
            float e = 20.0 * M_PI / 180.0f;
            loopi(19) {
                glNormal3f(0, 1, 0);
                glTexCoord3f(cos(-i * e) * radius / tdiv + dx, sin(-i * e) * radius / tdiv + dz, 0);
                glVertex3f(cos(-i * e) * radius, 0, sin(-i * e) * radius);
            } // loopi 19
            glEnd();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawUpperDome() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100, 100);

            int vsteps = 8;
            double vf = 0.5 * M_PI / (double) vsteps;

            int hsteps = 18;
            double hf = 2.0 * M_PI / (double) hsteps;
            double h = 0;

            float color1[3];
            memcpy(color1, mUpColor, sizeof (float) *3);

            float color2[3];
            memcpy(color2, mLoColor, sizeof (float) *3);

            float slope[3];
            set3i(slope, (color2[i] - color1[i]) / (float) vsteps);

            memcpy(color2, color1, sizeof (float) *3);
            vectoradd(color2, color2, slope);

            float s1 = sin(0);
            float c1 = cos(0);
            for (int i = 1; i <= vsteps; i++) {
                float s2 = sin(i * vf);
                float c2 = cos(i * vf);
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0; j <= hsteps; j++) {
                    float s = sin(h);
                    float c = cos(h);
                    glColor3fv(color1);
                    glNormal3f(s1*c, c1, s1 * s);
                    glVertex3f(s1*c, c1, s1 * s);
                    glColor3fv(color2);
                    glNormal3f(s2*c, c2, s2 * s);
                    glVertex3f(s2*c, c2, s2 * s);
                    h += hf;
                } // for hsteps
                glEnd();
                memcpy(color1, color2, sizeof (float) *3);
                vectoradd(color2, color2, slope);
                s1 = s2;
                c1 = c2;
            } // for vsteps
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawLowerDome() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100, 100);

            int vsteps = 8;
            double vf = 0.5f * M_PI / (double) vsteps;

            int hsteps = 18;
            double hf = 2.0 * M_PI / (double) hsteps;
            double h = 0;

            const float f = 0.5;
            float color2[3] = {0.4, 0.2, 0.1};
            memcpy(color2, mBtColor, sizeof (float) *3);
            float color1[3] = {color2[0] * f, color2[1] * f, color2[2] * f};

            float tmp[3];
            memcpy(tmp, color1, sizeof (float) *3);
            memcpy(color1, color2, sizeof (float) *3);
            memcpy(color2, tmp, sizeof (float) *3);

            float slope[3];
            set3i(slope, (color2[i] - color1[i]) / (float) vsteps);

            memcpy(color2, color1, sizeof (float) *3);
            vectoradd(color2, color2, slope);

            float s1 = -sin(0);
            float c1 = -cos(0);
            for (int i = 1; i <= vsteps; i++) {
                float s2 = -sin(i * vf);
                float c2 = -cos(i * vf);
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0; j <= hsteps; j++) {
                    float s = sin(h);
                    float c = cos(h);
                    glColor3fv(color2);
                    glNormal3f(s2*c, c2, s2 * s);
                    glVertex3f(s2*c, c2, s2 * s);
                    glColor3fv(color1);
                    glNormal3f(s1*c, c1, s1 * s);
                    glVertex3f(s1*c, c1, s1 * s);
                    h += hf;
                };
                glEnd();
                memcpy(color1, color2, sizeof (float) *3);
                vectoradd(color2, color2, slope);
                s1 = s2;
                c1 = c2;
            } // for vsteps
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawClouds() {
    // currently this method is a cpu waster.
    return; // !!!
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(1, 1, 1);
            glColor4f(0.9, 0.9, 0.95, 0.2);
            srand(1243);
            
            loopi(10) {
                float alpha = 0.017454f * (10 + (rand() % 75));
                float beta = 0.017454f * (rand() % 360);
                float x = sin(beta) * sin(alpha);
                float y = cos(alpha);
                float z = cos(beta) * sin(alpha);
                glPushMatrix();
                {
                    glTranslatef(30.0f * x, 30.0f * y, 30.0f * z);
                    glScalef(0.9, 0.6, 0.9);
                    glRotatef(90, 1,2,4);
                    int n = 3 + rand() % 6;
                    loopj(n) {
                        float a = 0.01 * (100 - (rand() % 200));
                        float b = 0.005 * (100 - (rand() % 200));
                        float c = 0.01 * (100 - (rand() % 200));
                        glTranslatef(a, b, c);
                        glCenterUnitBlock();
                        //glDisk(5,1);
                        //glCenterUnitCylinder(5);
                    } // loopj n
                }
                glPopMatrix();
            } // loopi 20
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawMountains() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glDisable(GL_CULL_FACE);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100, 100);

            unsigned int r = 1;
            loopi(6) {
                float f = 1.0f / (1.0f + 0.3 * i);
                int hsteps = 36 * i;
                float hf = 2 * M_PI / (float) hsteps;
                float h = 0;
                float u = 0;
                glBegin(GL_TRIANGLE_STRIP);
                glColor3f(0.3 * f, 0.3 * f, 0.4 * f);
                for (int j = 0; j <= hsteps; j++) {
                    r = (r * 1234567) & ((1UL << 16) - 1);
                    u += 0.2 * ((r % 100)*0.01 - 0.58);
                    if (u < 0) u = 0;
                    if (j == hsteps) u = 0;
                    u *= f;
                    float s = sin(h);
                    float c = cos(h);
                    glColor4f(0.6 * f, 0.65 * f, 0.7 * f, 0.95f);
                    glVertex3f(c, u, s);
                    glColor4f(0.1 * f, 0.3 * f, 0.25 * f, 1.0f);
                    glVertex3f(c, 0, s);
                    h += hf;
                } // for hsteps
                glEnd();
            } // loopi 6
        }
        glPopMatrix();
    }
    glPopAttrib();
}


