#include "cController.h"

#include "cObject.h"
#include "cWorld.h"

#include <cstdlib>
#include <cassert>

#define debug_state !true
#define debug_transitions true
#define debug_step !true

// -------------------------------------------------------------

cController::cController(cObject* entity, bool enabled) {
    mDevice = entity;
    mEnabled = enabled;
}

cController::~cController() {
    while (!mStack.empty()) mStack.pop_back();
}

const char* cController::getFrameName() {
    const char* names[] = {
        "WAIT", "ATTACK", "FOLLOW", "GOTO", "REPEAT"
    };
    unsigned int i = mStack.back();
    if (i >= sizeof(names)/sizeof(const char*)) return "NOSTATENAME";
    return names[i];
};

unsigned int cController::getFrameSizeOf(int opcode) {
    switch (opcode) {
        case WAIT: return 3;
        case ATTACK: return 2;
        case FOLLOW: return 3;
        case GOTO: return 5;
        case REPEAT : return 2;
        default: return 2;
    }
}

unsigned int cController::getFrameSize() {
    return getFrameSizeOf(mStack.back());
}

tSerial cController::getParameter(int offset) {
    return mStack[mStack.size() - offset - 1];
}

void cController::setParameter(int offset, tSerial value) {
    mStack[mStack.size() - offset - 1] = value;
}

void cController::push(tSerial value) {
    mStack.push_back(value);
}

void cController::pop() {
    if (debug_transitions) {
        printf("%s#%llu.pop()\n", mDevice->mName.c_str(), mDevice->mSerial);
    }
    int size = getFrameSize(); // Important: eval outside loop-condition!
    for (int i = 0; i < size; i++) mStack.pop_back();
    if (debug_transitions) {
        printf("Device: %s#%llu   %s   %i\n", mDevice->mName.c_str(), mDevice->mSerial, getFrameName(), getFrameSize());
    }
}

void cController::process() {
    if (debug_state) printf("cController::process()\n");

    if (!mEnabled) return;
    if (mDevice == NULL) return;

    if (mStack.empty()) pushWaitEvent();

    if (debug_step) printf("Device: %s#%llu  Frame: %s (%i)  StackSize: %i\n", mDevice->mName.c_str(), mDevice->mSerial, getFrameName(), getFrameSize(), mStack.size());

    switch (mStack.back()) {
        case WAIT: waitEvent();
            break;
        case ATTACK: attackEnemy();
            break;
        case FOLLOW: followLeader();
            break;
        case GOTO: gotoDestination();
            break;
        case REPEAT: repeatInstructions();
            break;
        default: printf("Invalid Instruction Request!\n");
            break;
    }
    if (debug_step) printf("-> Device: %s#%llu   %s   %i\n", mDevice->mName.c_str(), mDevice->mSerial, getFrameName(), getFrameSize());
}


// -------------------------------------------------------------

void cController::pushWaitEvent(long mseconds, bool patrol) {
    push(patrol);
    push(mseconds);
    push(WAIT);
}

void cController::waitEvent() {
    //tSerial opcode = getParameter(0);
    long mseconds = getParameter(1);
    bool patrol = getParameter(2);
    
    {
        mDevice->do_idle();
    }

    if (patrol) {
        tSerial nearby = mDevice->enemyNearby();
        if (nearby) {
            {
                tSerial self = mDevice->mSerial;
                cWorld::instance->sendMessage("#%llu: Intruder!\n", self);
            }
            pushAttackEnemy(nearby);
        }
    }

    if (mseconds > 0) {
        mseconds -= 1000/40;
        setParameter(1, mseconds);
        if (mseconds <= 0) {
            pop();
            return;
        }
    }
}

// -------------------------------------------------------------

void cController::pushAttackEnemy(tSerial entity) {
    if (debug_transitions) {
        printf("%s#%llu.pushAttackEnemy(%llu)\n", mDevice->mName.c_str(), mDevice->mSerial, entity);
    }
    {
        //tSerial self = mDevice->mSerial;
        //cWorld::instance->sendMessage("@%llu: All ur base belongs to us.\n", entity);
    }
    push(entity);
    push(ATTACK);
}
 
void cController::attackEnemy() {
    //tSerial opcode = getParameter(0);
    tSerial entity = getParameter(1);

    {
        mDevice->do_moveFor(NULL);
        mDevice->do_aimFor(entity);
        mDevice->do_moveNear();
        mDevice->do_aimAt();
        if (mDevice->inWeaponRange() > 0.5) mDevice->do_fireAt();
        //((cMech*)mDevice)->Pattern(tf, "nrnlln");
    }
    
    cObject* target = cWorld::instance->mIndex[entity];
    if (target == NULL) { // Target disappeared
        mDevice->do_aimFor(NULL);
        pop();
        return;
    } else if (target->hasRole(cObject::DEAD)) {
        mDevice->do_aimFor(NULL);
        pop();
        return;
    }

}

// -------------------------------------------------------------

void cController::pushFollowLeader(tSerial entity, bool patrol) {
    if (debug_transitions) {
        printf("%s#%llu.pushFollowLeader(%llu, %i)\n", mDevice->mName.c_str(), mDevice->mSerial, entity, patrol);
    }
    push(patrol ? 1 : 0);
    push(entity);
    push(FOLLOW);
}

void cController::followLeader() {
    //tSerial opcode = getParameter(0);
    tSerial entity = getParameter(1);
    tSerial patrol = getParameter(2);

    {
        mDevice->do_aimFor(entity);
        mDevice->do_moveNear();
    }

    if (patrol) {
        tSerial nearby = mDevice->enemyNearby();
        if (nearby) {
            pushAttackEnemy(nearby);
            return;
        }
    }
}

// -------------------------------------------------------------

void cController::pushGotoDestination(float* v, bool patrol) {
    if (v == NULL) return;
    if (debug_transitions) {
        printf("%s#%llu.pushGotoDestination((%f, %f, %f), %i)\n", mDevice->mName.c_str(), mDevice->mSerial, v[0],v[1],v[2], patrol);
    }
    unsigned long *p = (unsigned long*) v;
    push(patrol ? !0 : 0);
    push(p[2]);
    push(p[1]);
    push(p[0]);
    push(GOTO);
}

void cController::gotoDestination() {
    //tSerial opcode = getParameter(0);
    tSerial v0 = getParameter(1);
    tSerial v1 = getParameter(2);
    tSerial v2 = getParameter(3);
    tSerial patrol = getParameter(4);
    unsigned long p[] = {
        v0, v1, v2
    };
    float* v = (float*) p;

    {
        if (debug_state) printf("going %s to: %.2f %.2f %.2f\n", (patrol > 0) ? "patrolling" : "directly", v[0], v[1], v[2]);
        mDevice->do_moveFor(v);
        //mDevice->do_aimFor(0);
        mDevice->do_aimAt();
        mDevice->do_moveTowards();
    }

    float range = mDevice->inDestinationRange();
    if (debug_state) printf("DestinationRange %f\n", range);
    if (range > 0.0f) {
        pop();
        return;
    }

    if (patrol) {
        tSerial nearby = mDevice->enemyNearby();
        if (nearby) {
            pushAttackEnemy(nearby);
            return;
        }
    }
}

// -------------------------------------------------------------

void cController::pushRepeatInstructions(int n) {
    if (debug_transitions) {
        printf("%s#%llu.pushRepeatInstructions(%i)\n", mDevice->mName.c_str(), mDevice->mSerial, n);
    }
    push(n);
    push(REPEAT);
}

void cController::repeatInstructions() {
    tSerial opcode = getParameter(0);
    tSerial n = getParameter(1);

    if (debug_state) printf("%s  %i\n", getFrameName(), getFrameSize());

    unsigned int first = getFrameSize();
    if (debug_state) printf("first %i -> opcode %llu\n", first, opcode);
    unsigned int last = first;
    loopi(n) {
        int opcode = getParameter(last);
        last += getFrameSizeOf(opcode);
        if (debug_state) printf("last %i -> opcode %i\n", last, opcode);
    }

    int size = last - first;
    if (debug_state) printf("%i = %i - %i\n", size, last, first);

    for (int i = 0; i < size; i++) {
        push(getParameter(last - 1));
    }

    printf("################ REPEATING %i %i #################\n", (int) n, size);
}


