/**
 * File:   cController.h
 * Author: Benjamin Pickhardt (hackCraft.de)
 *
 * Created on March 28, 2008, 21:25 PM (1999)
 *
 * AI and autopilot stacked state machine for controlling entities.
 */

#ifndef _CONTROLLER_H
#define _CONTROLLER_H

class cController;

#include "cObject.h"

#include <vector>


/**
 * Autopilot/AI object controller, controlls usualy Mechs.
 * Implements an stack-automaton for object behavior controll
 * and plan execution.
 */
class cController {
public:
    /// Available opcodes of stack-machine.
    enum Opcodes {
        WAIT, ATTACK, FOLLOW, GOTO, REPEAT
    };

    bool mEnabled; // Whether control is enabled (autopilot/ai enabled/disabled).
    cObject* mDevice; // Controlled entity cObject.
    std::vector<tSerial> mStack;

public:
    /// Initialises a en-/disabled controller for the given entity.
    cController(cObject* entity, bool enabled = true);
    virtual ~cController();

    /// Do a single Instruction step as the top frame on the stack says.
    void process();

protected:
    const char* getFrameName();
    unsigned int getFrameSizeOf(int opcode);
    unsigned int getFrameSize();

    // Set Parameter value at top of the stack minus offset.
    void setParameter(int offset, tSerial value);

    // Get Parameter value from top of the stack minus offset.
    tSerial getParameter(int offset);

    // Push a single value onto the stack.
    void push(tSerial value);

    // Remove whole instruction/frame from stack.
    void pop();

public:
    // 1. Transition Function
    // 2. State Function

    // Wait until a Timeout occures or the selected event occures:
    // patrol = true  => Enemy Contact
    void pushWaitEvent(long mseconds = -1, bool patrol = true);
    void waitEvent();

    void pushAttackEnemy(tSerial entity);
    void attackEnemy();

    void pushFollowLeader(tSerial entity, bool patrol = true);
    void followLeader();

    void pushGotoDestination(float* v, bool patrol = true);
    void gotoDestination();

    void pushRepeatInstructions(int n);
    void repeatInstructions();

    // pushRandomWalk();

    // pushSendWave(tSerial entity, float radius, int mseconds, int attributes);

    // pushForValue(int n, int offset, int start, int end, int step);

    // Use these from controlled Object (mDevice):
    // Queries - ask for objects
    // Fuzzy Predicates [0,1] - ask for fulfillment
    // Output Actions - ask for taking action
};


#endif
