
#include <vector>

// cObject.hpp

#include "cObject.h"

#include "macros.h"

/**
 * This function collides a line given by start3fv and end3fv with an axis aligned box.
 * The box is given through two vectors - the minumum and maximum extends: min3fv and max3fv.
 * The function returns -1 if there is no collision and returns lamda between 0.0 and 1.0 otherwise.
 * The resulting impact point is copied into impactpoint3fv.
 * If there is an impact then the following is true: impactpoint3fv = start3fv + lamda * (end3fv - start3fv).
 * Therefore the distance from start3fv to impactpoint3fv equals: |(start3fv - end3fv)| * lamda.
 * The algo here itself is somewhat tricky sorry for missing information on that one.
 */
float collideLineWithAABB(float* start3fv, float* end3fv, float* min3fv, float* max3fv, float* impactpoint3fv, float* impactnormal3fv) {
    const float epsilon = 0.01f;

    float direction[6];
    direction[0] = direction[3] = end3fv[0] - start3fv[0]; // Calculate Direction vector of line - and replicate it to direction[3..5], too.
    direction[1] = direction[4] = end3fv[1] - start3fv[1]; // That replication is for easier access - no (i % 3) necessary.
    direction[2] = direction[5] = end3fv[2] - start3fv[2];

    float dists[6];
    memcpy(&dists[0], min3fv, sizeof (float) * 3);
    memcpy(&dists[3], max3fv, sizeof (float) * 3);
    const float nrmsign[] = {-1, -1, -1, +1, +1, +1};
#if 0
    printf("Checking Planes...\n");
#endif
    for (unsigned char i = 0; i < 6; i++) { // For every plane of bbox.
        if (nrmsign[i] * direction[i] > 0.0) continue; // Backfacing plane?
#if 0
        printf("Checking Frontfacing %i: %.2f\n", i, direction[i]);
#endif
        if ((direction[i] == +0.0f) || (direction[i] == -0.0f)) continue; // Line is parallel to this axis.
        float lamda = (dists[i] - start3fv[i % 3]) / (direction[i]); // Calculate lamda, so that (start + (dir * lamda)) = point_on_plane.
        if (lamda < 0.0f) continue; // Point is behind line origin point.
        if (lamda > 1.0f) continue; // Point is beyond line end point.
        float p[6];
        p[0] = p[3] = start3fv[0] + direction[0] * lamda; // Calculate point on plane using previous lamda.
        p[1] = p[4] = start3fv[1] + direction[1] * lamda;
        p[2] = p[5] = start3fv[2] + direction[2] * lamda;
        int j;
        for (j = 0; j < 6; j++) { // For every plane of bbox - check if point inside.
            if (j == i) {
                continue;
            } else if (j < 3) {
                if (p[j] < dists[j] - epsilon) break; // Not Inside if less than min3fv.
            } else {
                if (p[j] > dists[j] + epsilon) break; // Not Inside if greater than max3fv.
            }
        }
        if (j == 6) { // then previous test passed.
            memcpy(impactpoint3fv, p, sizeof (float) * 3);
            memset(impactnormal3fv, 0, sizeof (float) * 3);
            impactnormal3fv[i % 3] = nrmsign[i];
            return lamda;
        }
    }
    return -1.0f;
}



#define set3f(v, x,y,z) { (v)[0] = x; (v)[1] = y; (v)[2] = z; }

#define cpy3f(vdest, vsrc) memcpy(vdest, vsrc, sizeof(float)*3)

cObject::cObject() {
    //printf("cObject::cObject()\n");

    // BASE
    mSerial = 0; // Not yet spawned, no spawndate
    mRoles = 0;
    mSeconds = 0;

    // NAMEABLE
    mName = "Anonymous";
    mDescription = "Undescribed";
    mDesignation = 0;

    // SOCIAL
    mAllies = 0;
    mEnemies = 0;

    // TRACKABLE
    mAcc.reserve(3);
    mAcc.resize(3);
    mVel.reserve(3);
    mVel.resize(3);
    mPos.reserve(3);
    mPos.resize(3);
    mRot.reserve(3);
    mRot.resize(3);
    set3f(mAcc, 0.0f, 0.0f, 0.0f);
    set3f(mVel, 0.0f, 0.0f, 0.0f);
    set3f(mPos, 0.0f, 0.0f, 0.0f);
    set3f(mRot, 0.0f, 0.0f, 0.0f);
    mRadius = 1;

    // DAMAGEABLE
    loopi(MAX_PARTS) {
        mHp[i] = 100.0f;
    }
    //mSplasharmor = NULL;
    //mBeamarmor = NULL;
    //mPiercearmor = NULL;

    // ENTITY
    mTarget = 0;
    mDestination.reserve(3);
    mDestination.resize(3);
    set3f(mDestination, 0.0f, 0.0f, 0.0f);
    mPad = NULL;
    mController = NULL;
}

int cObject::ENABLE_TEXTURE_3D = 1;

cObject::cObject(cObject* source) {
    //printf("cObject::cObject(cObject*)\n");

    // BASE
    mRoles = source->mRoles;
    mSeconds = source->mSeconds;

    // NAMEABLE
    mName = source->mName;
    mDesignation = source->mDesignation;

    // SOCIAL
    mAllies = source->mAllies;
    mEnemies = source->mEnemies;

    // TRACKABLE
    mAcc = source->mAcc;
    mVel = source->mVel;
    mPos = source->mPos;
    mRot = source->mRot;
    mRadius = source->mRadius;

    // DAMAGEABLE

    loopi(MAX_PARTS) {
        mHp[i] = source->mHp[i];
    }
    //mSplasharmor = source->mSplasharmor;
    //mBeamarmor = source->mBeamarmor;
    //mPiercearmor = source->mPiercearmor;

    // ENTITY
    mTarget = source->mTarget;
    mDestination = source->mDestination;
    //cpy3f(mDestination, source->mDestination);
    mPad = source->mPad;
    mController = source->mController;
}

float cObject::checkRelHeight() {
    float* g = cWorld::instance->mGravity.data();

    float len = 10;
    float a[] = {mPos[0], mPos[1] + 2, mPos[2]};
    float b[] = {a[0] - g[0] * len, a[1] - g[1] * len, a[2] - g[2] * len};

    float nearest2 = 1000;
    int n = 0;

    unsigned long roles = (1UL << cObject::COLLIDEABLE);
    float beamradius = 1;
    float rangeradius = 35;

    std::list<cObject*>* range = cWorld::instance->filterByRange(this, a, 0.0f, rangeradius, -1);
    if (!range->empty()) {
        //printf("range: \n"); cWorld::instance->printNames(range);
        //printf("range %i\n", range->size());
        std::list<cObject*>* role = cWorld::instance->filterByRole(this, roles, true, -1, range);
        if (!role->empty()) {
            //printf("roles: \n"); cWorld::instance->printNames(role);
            //printf("roles %i\n", role->size());

            looplist(i, cObject*, *role) {
                // todo: min dist
                cObject::rImpact* imp = ((cObject*) (*i))->damageByBeam(a, b, beamradius, 0, true);
                if (imp != NULL) {
                    float v[] = {imp->pos[0] - a[0], imp->pos[1] - a[1], imp->pos[2] - a[2]};
                    float d2 = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
                    if (d2 < nearest2) nearest2 = d2;
                    n++;
                    delete imp;
                }
            }
        }
        delete role;
    }
    delete range;

    if (n > 0) {
        //printf("below: %i, %.2f\n", n, sqrt(nearest2));
        return sqrt(nearest2) - 2;
    }

    return mPos[Y];
}

float cObject::checkAhead(float seconds) {
    // Forward Movement
    //const float e = 0.017454;

    float* dir = mVel.data();
    float velocity = sqrt(mVel[X] * mVel[X] + mVel[Y] * mVel[Y] + mVel[Z] * mVel[Z]);

    float len = 10;
    float a[] = {mPos[0], mPos[1] + 2, mPos[2]};
    float b[] = {a[0] + dir[0] * len, a[1] + dir[1] * len, a[2] + dir[2] * len};

    float nearest2 = len*len;
    int n = 0;

    unsigned long roles = (1UL << cObject::COLLIDEABLE);
    float beamradius = 1;
    float rangeradius = 35;

    // Check against groundplane while descending.
    if (mVel[Y] < 0.0f) {
        //printf("%.3f   %.3f\n", mPos[Y], mVel[Y]);
        float d = a[Y] / -mVel[Y];
        float v[] = {d * mVel[X], d * mVel[Y], d * mVel[Z]};
        float d2 = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
        if (d2 < nearest2) nearest2 = d2;
    }

    std::list<cObject*>* range = cWorld::instance->filterByRange(this, a, 0.0f, rangeradius, -1, NULL);
    if (!range->empty()) {
        //printf("range: \n"); cWorld::instance->printNames(range);
        //printf("range %i\n", range->size());
        std::list<cObject*>* role = cWorld::instance->filterByRole(this, roles, true, -1, range);
        if (!role->empty()) {
            //printf("roles: \n"); cWorld::instance->printNames(role);
            //printf("roles %i\n", role->size());

            looplist(i, cObject*, *role) {
                // todo: min dist
                cObject::rImpact* imp = ((cObject*) (*i))->damageByBeam(a, b, beamradius, 0, true);
                if (imp != NULL) {
                    float v[] = {imp->pos[0] - a[0], imp->pos[1] - a[1], imp->pos[2] - a[2]};
                    float d2 = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
                    //printf("ahead: %.2f\n", d2);
                    if (d2 < nearest2) nearest2 = d2;
                    n++;
                    delete imp;
                }
            }
        }
        delete role;
    }
    delete range;

    float nearest = sqrt(nearest2);

    float vt = velocity * seconds;

    float d = nearest - mRadius;

    //printf("%i ahead: %.3f  ahead-radius: %.3f  vel: %.3f  veltime: %.3f\n", n, nearest, d, v, vt);
    if (vt > d) {
        float f = d / vt;

        loopi(3) {
            mVel[i] *= f;
        }
    }

    return nearest;
}

