/**
 * File:   cWorld.h
 * Project: LinWarrior 3D
 *
 * Created on March 28, 2008, 10:23 PM (1999)
 */

#ifndef _CWORLD_H
#define _CWORLD_H

/**
 * Serial-Code or OID for cObject (tSerial is used for values, too).
 * Use whereever possible to store object references.
 * Object A having a pointer to cObject B is like saying A and B are bound together.
 * Pointers are ok between a Mech and its *own* weapon but not between
 * a targeting weapon (or Mech) and a hostile target Mech.
 * Only bind together what belongs together and what would be destroyed together.
 * Pointers express responsibility whereas OIDs loosely couple.
 * Use cWorld::instance->mIndex[serial] to map serial to pointer.
 */
typedef unsigned long long tSerial;

class cWorld;
#include "psi3d/macros.h"
#include "cObject.h"

#include <list>
#include <map>
#include <string>
#include <vector>

#include <kdtree.h>

class cBackground;
class cObject;
class cMission;

/**
 * The Game-Object-Managing World-Database-Class.
 * 
 */
class cWorld {
public:

    struct rMessage {
        tSerial mSerial;
        //unsigned long mReceiverRoles;
        //long mSeconds;
        std::string mText;
    };

    std::list<rMessage*> mMessages;

    /// The current World instance (singleton) other objects may
    /// only affect that current instance.
    static cWorld* instance;

    /// Background Object
    cBackground* mBackground;

    // Not Yet, may be merged with Background:
    // Overlays weather effects like rain, snow or dust.
    //cWeather* mWeather;

    /// Mission Setup and Objective-Controller.
    cMission* mMission;

    /// Downward force vector (or 0,0,0 in space) -
    /// kind of featuritis, a scalar may be enough.
    std::vector<float> mGravity;

    /// Object-ID (OID) to Object-Reference (non-fragged).
    std::map<tSerial, cObject*> mIndex;

    /// Contains non-fragged objects.
    std::list<cObject*> mObjects;

    /// Contains fragged objects.
    std::list<cObject*> mCorpses;

    kdtree<cObject*>* mCluster;

    /*
     * Key-Value-Pairs for world state and progress indication.
     * Interpretation is not fixed and may vary, samples:
     * "ImportantThingHappened" = "true" [boolean value]
     * "CurrentChapter" = "1" [integer value]
     * "ObjectOfInterest" = "20280809120000001" [tSerial for cObject]
     * "JustAFlagOrKeyitem" = "" [flag]
     */
    //std::map<std::string, std::string> mValues;

    // Serious question: What time & date format would be used
    // as a common planetary calendar? Is that sensible at all?!
    // So far I assume terran planetary dates everywhere.
    // But time & date should be adjusted to the planet and
    // planet's date shall be mapped like beween maya, gregorian
    // or moon calendars.

    /// Counts years of simulation universe.
    unsigned int mYear;

    /// Counts 366-Days of simulation year (0-365).
    unsigned int mDay;

    /// Counts 24-Hours of simulation day (0-23).
    unsigned int mHour;

    /// Counts minutes of simulation hour.
    unsigned int mMinute;

    /// Counts seconds of simulation minute.
    unsigned int mSecond;

    /// Micro-Seconds, incremented at every simulation timestep
    unsigned int mMSec;

    /// Incremented at every simulation timestep and reset on each second.
    unsigned int mFrame;

    /// Incremented at every (noteable) event within a frame. Reset each frame.
    unsigned int mDeltacycle;

    /// seconds per frame = 1.0f/mFPS.
    float mSPF;

    /// frames per second = 1.0f/mSPF).
    float mFPS;

public: // Constructor, Object-Management and Drawing:

    cWorld();

public: // Querries

    tSerial getSerial();

public: // Messaging

    void sendMessage(const char* format, ...);
    void sendMessageT(const char* text);

public: // Spawning And Fragging

    /**
     * Spawns object immediately, assigns serial OID and
     * finally calls object->onSpawn().
     */
    void spawnObject(cObject *object);

    /**
     * Frags an object by removing from active world and adding to Corpse-List.
     * Adds cObject::FRAGGED role to object, removes it from mObjects-List
     * and from mIndex-Map and inserts it into the mCorpses-List
     * so it can be deleted later on by bagFragged().
     * Finally object->onFragged() is called.
     */
    void fragObject(cObject *object);

    /**
     * Deletes fragged objects from Corpse-List.
     * On frag Objects aren't deleted instantly.
     * Instead Objects reside inside a Graveyard-list
     * and are finally removed and deleted by this call.
     * Possibly called very rarely within a Mission at transitional moments.
     */
    void bagFragged();

public: // Simulation Step - Call every frame in order to update the world/mission and draw a frame.

    /// Advance simulation time for one frame.
    void advanceTime(int deltamsec);
    /// Let all objects process input, adjust pose and calculate physics.
    void animateObjects();
    /// Let all objects calculate transformation matrices etc.
    void transformObjects();
    /// Draws background (skybox) by calling mBackground drawing method.
    void drawBack();
    /// Draw all Object's solid surfaces (calls their drawSolid method).
    void drawSolid(cObject* camera);
    /// Draw all Object's effects (calls their drawEffect method).
    void drawEffect(cObject* camera);

public: // World-Filtering, World-Scanning, World-Sense for objects.

    /**
     * Debug function to print out an object-list.
     * With a parameter of NULL/no parameter at all prints all objects within the world.
     * Warning: Printing all objects could lead to massive load on the console.
     */
    void printNames(std::list<cObject*>* objects = NULL);

    /**
     * Filters the given objectlist (or all mObjects if NULL) by roles (ORed Bitmask) (excluding ex Object).
     * There is the option to match all roles (true=AND) or to select any given role (false=OR).
     * The returned list is fresh allocated - caller delete responsibility.
     */
    std::list<cObject*>* filterByRole(cObject* ex, unsigned long rolemask, bool all, int maxamount, std::list<cObject*>* objects = NULL);

    /** Returns a List of objects which are within minimum and maximum range (excluding ex Object).
     * The returned list is fresh allocated - caller delete responsibility.
     */
    std::list<cObject*>* filterByRange(cObject* ex, float* origin, float minrange, float maxrange, int maxamount, std::list<cObject*>* objects = NULL);

    /**
     * Returns a List of objects which have the specified name.
     * The returned list is fresh allocated - caller delete responsibility.
     */
    std::list<cObject*>* filterByName(cObject* ex, char* name, int maxamount, std::list<cObject*>* objectlist = NULL);

    /**
     * Returns a List of objects whose bounding volume (approximately) intersects with the cylinder
     * described by the volume of certain diameter around a line.
     * The returned list is fresh allocated - caller delete responsibility.
     */
    std::list<cObject*>* filterByBeam(cObject* ex, float* pointa, float* pointb, float radius, int maxamount, std::list<cObject*>* objects = NULL);
};

#endif
